/* Author: Domagoj Babic 2004.
 *
 * $Id: queues.h,v 1.8 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#ifndef H_QUEUES
#define H_QUEUES

#include <string.h>
#include "utils.h"
#include "sort.h"

typedef uint Item;

// **************** INFLATABLE STACK *****************

#define STACKMOVE_TRESHOLD 2

typedef struct stack_t {
    uint size;
    uint idx;
    Item *a;
} stack_t;

stack_t *initStack(const uint size) {
    uint N = size + 1;
    stack_t *s = xmalloc(sizeof(stack_t));
    s->a = xmalloc(N * sizeof(Item));
    s->size = N;
    s->idx = 0;
    return s;
}

inline uint stackEmpty(stack_t *q) {
    return q->idx == (uint)0;
}

inline void stackReset(stack_t *q) {
    q->idx = 0;    
}

inline uint stackSize(stack_t *s) {
    return s->idx;
}

/*
inline void setStackPtr(stack_t *s, uint idx) {
    assert(idx <= s->size);
    s->idx = idx;
}// */

inline void stackPush(stack_t *q, const Item x) {
    if (q->idx >= q->size) {
        (q->size) <<= 1;
        q->a = (Item *)xrealloc((char *)q->a, q->size*sizeof(Item));
    }
    q->a[(q->idx)++] = x;
}

/*
inline void stackPushNC(stack_t *s, const Item x) {
    assert(s->idx < s->size);
    s->a[(s->idx)++] = x;
}

inline Item stackRead(stack_t *q, int offset) {
    offset++;
    assert((uint)offset < q->idx && "Invalid offset!");
    return q->a[q->idx - offset];
}// */

inline Item stackReadAt(stack_t *s, int i) {
    assert((uint)i < s->idx);
    return s->a[i];
}

inline void stackWriteAt(stack_t *s, int i, Item dat) {
    assert((uint)i < s->size);
    s->a[i] = dat;
}

void stackFree(stack_t *s) {
    free(s->a);
    free(s);
}

/*
void stackDeflate(stack_t *s) {
    char deflate = FALSE;
    while (s->idx < (s->size >> 2)) {
        s->size >>= 1;
        deflate = TRUE;
    }
    if (deflate) {
        Item *tmp = xmalloc(s->size * sizeof(Item));
        memcpy(tmp, s->a, s->idx * sizeof(Item));
        free(s->a);
        s->a = tmp;
    }
}// */

inline Item stackPop(stack_t *q) {
    DBG(assert(!stackEmpty(q) && "Stack underflow!");)
    return q->a[--(q->idx)];
}

/*
inline void stackQSort(stack_t *q) {
    cmp_qsort(q->a, q->idx, 4, ucmp);
}

inline int stackSearch(stack_t *q, uint x) {
    return binsearch((int *)q->a, (int)q->idx, (int)x);
}

inline void stackExpand(stack_t *s, uint sz) {
    assert(stackEmpty(s));
    if (s->size < sz) {
        s->a = (Item *)xrealloc((char *)s->a, sz * sizeof(Item));
        s->size = sz;
        assert(s->idx == 0);
    }
}// */

inline int findInStack(stack_t *s, Item x) {
    int i;
    for (i = 0; (uint)i < s->idx; i++) {
        if (s->a[i] == x) {
            return i;
        }
    }
    return -1;    
}

inline void stackDelEl(stack_t *s, uint i) {
    assert(i < s->idx);
    assert(s->idx > 0);
    swap((int *)s->a, (int)i, (int)(s->idx - 1));
    s->idx--;
}

/*
inline void stackCopy(stack_t *d, stack_t *s) {
    char inflate = FALSE;
    while (d->size < s->idx + d->idx) {
        d->size <<= 1;
        inflate = TRUE;
    }
    if (inflate) {
        d->a = (Item *)xrealloc((char *)d->a, d->size*sizeof(Item));
    }
    memcpy(&(d->a[d->idx]), s->a, s->idx*sizeof(Item));
    d->idx += s->idx;
}

inline void stackCopyNC(stack_t *d, stack_t *s) {
    assert(d->size >= s->idx + d->idx);
    memcpy(&(d->a[d->idx]), s->a, s->idx*sizeof(Item));
    d->idx += s->idx;
}

inline void stackMove(stack_t *d, stack_t *s) {
    stackCopy(d, s);
    stackReset(s);
}// */

inline void stackDump(stack_t *s, uint **arr, uint *size) {
    uint idx = s->idx;
    *size = idx;
    if (idx) {
        uint *dmp = xmalloc(idx * sizeof(uint));
        memcpy(dmp, s->a, idx * sizeof(uint));
        *arr = dmp;
    }
}

/*
inline void stackFillSzNC(stack_t *s, uint *a, uint sz) {
    if (!sz) {
        return;
    }
    assert(s->idx + sz <= s->size);
    memcpy(&(s->a[s->idx]), a, sz * sizeof(uint));
    s->idx += sz;
}// */

void reverseStack(stack_t *s) {
    uint n = stackEmpty(s) ? 0 : stackSize(s) - 1;
    uint i = 0;
    while (n && n > i) {
        swap((int *)s->a, (int)i, (int)n);
        i++;
        n--;
    }
}

inline void stackFillNC(stack_t *s, uint *start, uint *end) {
    uint n = (uint)(end - start);
    assert(start <= end);
    assert(s->idx + n <= s->size);
    if (n) {
        memcpy(&(s->a[s->idx]), start, n * sizeof(uint));
        s->idx += n;
    }
}

/* no checking whether d can receive content from s
inline void stackMoveNC(stack_t *d, stack_t *s) {
    stackCopyNC(d, s);
    stackReset(s);    
}// */

/* **************** INFLATABLE FIFO ******************

typedef struct fifo_t {
    uint size;
    uint head;
    uint idx;
    Item *a;
} fifo_t;

fifo_t *initFifo(const uint size) {
    uint N = size + 1;
    fifo_t *f = xmalloc(sizeof(fifo_t));
    f->a = xmalloc(N * sizeof(Item));
    f->size = N;
    f->head = N;
    f->idx = 0;
    return f;
}

inline void fifoReset(fifo_t *q) {
    q->head = q->size;
    q->idx = 0;
}
 
inline int fifoEmpty(fifo_t *q) {
    return q->head % q->size == q->idx;
}

inline uint fifoSize(fifo_t *q) {
    if (q->idx >= q->head) {
        return q->idx - q->head;
    } else {
        return q->size - q->head + q->idx;
    }
}

inline void fifoFree(fifo_t *q) {
    free(q->a);
    free(q);
}

void fifoPut(fifo_t *q, const Item x) {
    if (q->head % q->size == (q->idx + 1) % q->size) { 
        // inflate fifo
        uint oldSize = q->size, newSize = q->size << 1;
        q->a = (Item *)xrealloc((char *)q->a, newSize*sizeof(Item));
        memcpy(&(q->a[oldSize]), q->a, oldSize*sizeof(Item));
        q->head += oldSize;
        q->size = newSize;
    }
    q->a[(q->idx)++] = x;
    q->idx %= q->size;
}

inline Item fifoGet(fifo_t *q) {
    q->head %= q->size;
    return q->a[(q->head)++];
}// */

#endif // H_QUEUES

/*
 * Revision history:
 *
 * $Log: queues.h,v $
 * Revision 1.8  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.7  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.5  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.4  2004/12/11 04:46:12  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.3  2004/10/31 19:11:34  root
 * Added a couple of new stack functions.
 *
 * Revision 1.2  2004/10/24 00:07:21  root
 * Added stack move and copy functions.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */

