/* Author: Domagoj Babic 2004.
 *
 * $Id: simplify.h,v 1.9 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#include <ctype.h>
#include <stdlib.h>
#include "datas.h"
#include "graph.h"

#ifndef H_SIMPLIFY
#define H_SIMPLIFY

// EQR_UNIT_OR_BIN-> unit or bin clause generated
// EQR_EMPTY      -> either a conflict or redundant cl, c2 is deleted
// EQR_REGULAR    -> normal reduction                 
enum reduction_typ_e {EQR_UNIT_OR_BIN, EQR_EMPTY, EQR_REGULAR};

enum eqv_var_typ_e {EQ_NONE = 0, EQ_TAUTOLOGY, EQ_DEPENDANT};

enum pp_stats_e {PP_EVNT_BINEQ, PP_EVNT_TAUT, PP_EVNT_NUMEQ,
    PP_EVNT_UNITV, PP_EVNT_NUM_ELIM_VAR, PP_EVNT_NUM_ELIM_CLS,
    PP_EVNT_PURELIT, PP_TIME_GRAPHGEN, PP_TIME_CLUSTER};

typedef struct preproc_t {
    uint ovnum; // number of variables in original CNF
    uint min_clu_perc;
    uint *tdif; // temporary, clause dif
    uint *tnlt; // temporary, lits added from 2nd clause in clause dif
    uint *tins; // temporary, intersection of two clauses
    uint **pri;     // priority matrix
    bcpq_t *impq;
    stack_t *taut_stack;
    al_grph_t *eq_graph;
    // x = model[i], 1 <= i <= vnum:
    // x = 0 -> i FALSE
    // x = 1 -> i TRUE
    // x > 1 -> i <=>  getVar(x) if LSB(x)==1
    //       -> i !<=> getVar(x) if LSB(x)==0
    // x = UINT_MAX -> i is a tautology literal
    uint *model;    // partial assignment
    // equivalence clauses
    // x = eqs[i], 0 <= i < numer of CNF clauses
    // x = NULL - clause was zero-sized and deleted after preprocessing
    // x = cla_t *cl
    //      getClSize(cl) = 0 -> clause was deleted (unit, binary,
    //      and redundant equivalence clauses get deleted)
    //      getClSize(cl) > 1 -> plain eqv clause, tautology, or
    //      a clause with dependant var:
    //          isTautology(cl) = TRUE -> tautology
    //          isTautology(cl) = FALSE -> plain or clause with
    //          a dependant var:
    //              cl[clsz] = 0 -> plain clause
    //              cl[clsz] > 0 -> var cl[clsz] is a dependant var
    cla_t **eqs;
    // pointer where the next equivalence clause will be written
    uint eqs_ptr;
    // dependant variables
    flexarray_t *depvars;
    // clusters
    flexarray_t *clusters;
    // elimvars contains replaced, tautology, and eliminated unit vars
    flexarray_t *elimvars;
    flexarray_t *ocnf;      // original CNF formula
    flexarray_t *ecnf;
    // entries in clst are clause pointers, as clauses never increase
    // their size, hence there's no need to change the pointers all
    // around
    flexarray_t **clst;     // table of clause arrays, indexed by lit
    // when the difference of two equivalencies is computed, some
    // variables remain in the result, some don't and some new are
    // added. to avoid updating the pointer in equivalence table
    // (elst) for variables that remain the same, the table contains
    // indices into eqs, rather than pointers.
    flexarray_t **elst;     // table of equivalencies, indexed by var
    stats_t *stats;
    uint bin_boost;
    uint sat_status; 
    char eqvPreprocessing;
    char clusterize;
} preproc_t;

// **************** DECLARATIONS **********************

flexarray_t *parse(char *name, uint *vnum, uint *clnum);
int appendInt(flexarray_t *cls, int x, flexarray_t *f);
void freeClLst(flexarray_t *fl);
#ifdef DEBUG
void printCNF(preproc_t *p, char *ofile);
#endif
void populate_aux(preproc_t *p, flexarray_t *formula, flexarray_t **tbl);
preproc_t *initPreproc(uint vnum, flexarray_t *clst, stats_t *stats, 
        char eqvPreprocessing, char clusterize, uint min_cl_perc);
void freePreproc(preproc_t *p);
void partiallyFreePreproc(preproc_t *p);
int clcmp(const void *ca, const void *cb);
void addcl_coe_aux(preproc_t *p, cla_t *c);
void addcl_cnf_aux(preproc_t *p, cla_t *c);
void delone_aux(flexarray_t **tbl, uint idx, uint dat);
void delcl_cnf_aux(preproc_t *p, cla_t *c);
void delcl_coe_aux(preproc_t *p, cla_t *c, uint cidx);
void delset_cnf_aux(preproc_t *p, flexarray_t **tbl, uint lit);
void dellit_aux(preproc_t *p, uint lit, cla_t **eqs);
void repset_aux(preproc_t *p, uint l1, uint l2);
void elimUnitCl(preproc_t *p, cla_t *c, int eqv_idx);
void repset_eqv_aux(preproc_t *p, uint v1, uint v2, uint eqv);
char impqPut(bcpq_t *iq, uint lit);
uint impqGet(bcpq_t *iq);
char bcp(preproc_t *p);
char propagate_bieq(preproc_t *p);
char onestep_resolve(preproc_t *p);
char eq_find(preproc_t *p);
char eq_reason(preproc_t *p);
char eq_reduce(preproc_t *p, uint c1i, uint c2i, char onlyIfShorter);
char eq_find_depvars(preproc_t *p);
void simplify(preproc_t *p);
void verify(preproc_t *p, char *instance);
void postprocess(preproc_t *p, flexarray_t *model, char format,
        char *ofile, char *instance, clock_t start);
void elimTautology(preproc_t *p, cla_t *c, uint cl_idx);
inline uint occurInCNF(preproc_t *p, uint v);
inline uint occurInCOE(preproc_t *p, uint v);
void resolveTautEqvCl(preproc_t *p, cla_t *c);
uint resolveBinEq(preproc_t *p, uint v);
//void setECNF(preproc_t *p);
#ifdef COLLECT_STATS
void reportStats(stats_t *s, char *ofile);
#endif
//void statsUpdate(stats_t *s, char event, uint x);
char find_bieq(preproc_t *p);

#ifdef COLLECT_STATS
void statsUpdate(stats_t *s, char event, long x) {
    switch(event) {
        case PP_EVNT_BINEQ:
            s->ppBinEqs++;
            break;
        case PP_EVNT_TAUT:
            s->ppTautologies++;
            break;
        case PP_EVNT_NUMEQ:
            s->ppNumEqs = x;
            break;
        case PP_EVNT_UNITV:
            s->ppUnitVars++;
            break;
        case PP_EVNT_NUM_ELIM_VAR:
            s->ppTotVarElim = x;
            break;
        case PP_EVNT_NUM_ELIM_CLS:
            s->ppTotClsElim++;
            break;
        case PP_EVNT_PURELIT:
            s->ppPureLit++;
            break;
        case PP_TIME_CLUSTER:
            s->ppTimeCluster = (double)x / CLOCKS_PER_SEC;
            break;
        case PP_TIME_GRAPHGEN:
            s->ppTimeGraphGen = (double)x / CLOCKS_PER_SEC;
            break;
    }
}
#else
#define statsUpdate(a,b,c)
#endif

#ifdef COLLECT_STATS
void reportStats(stats_t *s, char *ofile) {
    FILE *fp = NULL;
    char *msg;
    msg = xmalloc(256);
    if (ofile) {
        fp = fopen(ofile, "a+");
        if (!fp) error("Output file creation error!");
    }
    (void)computeStats(s);
    sprintf(msg, "c ******************* STATISTICS ********************\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> Preprocessing statistics <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Binary equivalences: %ld\n", s->ppBinEqs);
    writeStream(msg, fp);
    sprintf(msg, "c Tautologies: %ld\n", s->ppTautologies);
    writeStream(msg, fp);
    sprintf(msg, "c Other equivalences: %ld\n", s->ppNumEqs);
    writeStream(msg, fp);
    sprintf(msg, "c Unit variables: %ld\n", s->ppUnitVars);
    writeStream(msg, fp);
    sprintf(msg, "c Pure literals: %ld\n", s->ppPureLit);
    writeStream(msg, fp);
    sprintf(msg, "c Binary clauses in simplified problem: %ld\n", s->ppNumBinCls);
    writeStream(msg, fp);
    sprintf(msg, "c Total eliminated variables: %ld\n", s->ppTotVarElim);
    writeStream(msg, fp);
    sprintf(msg, "c Total eliminated clauses: %ld\n", s->ppTotClsElim);
    writeStream(msg, fp);
    sprintf(msg, "c Number of clusters: %u\n", s->ppNumClusters);
    writeStream(msg, fp);
    sprintf(msg, "c TIME:graph generation: %f [s]\n", s->ppTimeGraphGen);
    writeStream(msg, fp);
    sprintf(msg, "c TIME:clustering: %f [s]\n", s->ppTimeCluster);
    writeStream(msg, fp);
    sprintf(msg, "c\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> Clause statistics <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Initial cache size: %u\n", s->cacheSize);
    writeStream(msg, fp);
    sprintf(msg, "c Cache cleanup percentage: %u\n", s->cacheDelPerc);
    writeStream(msg, fp);
    sprintf(msg, "c Cache inflation rate: %f\n", s->cacheInflationRate);
    writeStream(msg, fp);
    sprintf(msg, "c Original cls: %u\n", s->clOrigClauses);
    writeStream(msg, fp);
    sprintf(msg, "c\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> Decisions <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Variables: %u\n", s->varNum);
    writeStream(msg, fp);
    sprintf(msg, "c Decisions: %ld\n", s->decNum);
    writeStream(msg, fp);
    sprintf(msg, "c Fixed through learning: %u\n", s->decNumFixedLearn);
    writeStream(msg, fp);
    sprintf(msg, "c Max dec lev reached: %u\n", s->decMaxLev);
    writeStream(msg, fp);
    sprintf(msg, "c Max num of assigned vars: %u\n", s->decMaxAssigned);
    writeStream(msg, fp);
    sprintf(msg, "c\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> B-cubing <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Branches per # of flips: %f\n", s->hcFlipBranchDensity);
    writeStream(msg, fp);
    sprintf(msg, "c Nodes per # of flips: %f\n", s->hcFlipNodeDensity);
    writeStream(msg, fp);
    sprintf(msg, "c Average # of branches: %f\n", s->hcAvgBranches);
    writeStream(msg, fp);
    sprintf(msg, "c Average # of nodes: %f\n", s->hcAvgNodes);
    writeStream(msg, fp);
    sprintf(msg, "c Average # of used branches: %f\n", s->hcAvgUsedBranches);
    writeStream(msg, fp);
    sprintf(msg, "c Average # of used nodes: %f\n", s->hcAvgUsedNodes);
    writeStream(msg, fp);
    sprintf(msg, "c Non empty B-cubes: %ld\n", s->hcNumHcubes);
    writeStream(msg, fp);
    sprintf(msg, "c Intersections: %ld\n", s->hcIntersections);
    writeStream(msg, fp);
    sprintf(msg, "c Intersection conflicts: %ld\n", s->hcIntersectConflicts);
    writeStream(msg, fp);
    sprintf(msg, "c\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> Implications <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Total implications: %ld\n", s->impNumImps);
    writeStream(msg, fp);
    sprintf(msg, "c Implication chains: %ld\n", s->impNumChains);
    writeStream(msg, fp);
    sprintf(msg, "c Average imp chain length: %f\n", s->impAvgChainLen);
    writeStream(msg, fp);
    sprintf(msg, "c Longest imp chain: %u\n", s->impLongestChain);
    writeStream(msg, fp);
    sprintf(msg, "c Scheduled imp blocks: %ld\n", s->impBlocksScheduled);
    writeStream(msg, fp);
    sprintf(msg, "c Average scheduled block size: %f\n", 
            s->impAvgSchBlockSz);
    writeStream(msg, fp);
    sprintf(msg, "c\n");
    writeStream(msg, fp);
    sprintf(msg, "c -> Conflicts <-\n");
    writeStream(msg, fp);
    sprintf(msg, "c Conflicts: %ld\n", s->cnfNum);
    writeStream(msg, fp);
    sprintf(msg, "c Average DEC clause size: %f\n", s->cnfDECAvgSz);
    writeStream(msg, fp);
    sprintf(msg, "c Total learned cls: %ld\n", s->cnfUIPs);
    writeStream(msg, fp);
    sprintf(msg, "c Average learned clause size: %f\n", s->cnfUIPAvgSz);
    writeStream(msg, fp);
    sprintf(msg, "c Restarts: %u\n", s->restarts);
    writeStream(msg, fp);
    sprintf(msg, "c ***************************************************\n");
    writeStream(msg, fp);
    if (ofile) {
        fclose(fp);
    }
    free(msg);
}
#endif

void freeClLst(flexarray_t *fl) {
	uint sz, *a;
    assert(fl != NULL);
    sz = getFlexArraySize(fl);
    a = getFlexArray(fl);
    while (sz > 0) { 
        cla_t *c = (cla_t *)a[--sz];
        free(c);
    }
    resetFlexArray(fl);
}

preproc_t *initPreproc(uint vnum, flexarray_t *clst, stats_t *stats, 
        char eqvPreprocessing, char clusterize, uint min_cl_perc) {
    uint i, lits = (vnum + 1) << 1;
    uint clnum = getFlexArraySize(clst);
    preproc_t *p = xmalloc(sizeof(preproc_t));
    p->stats = stats;
    
    p->sat_status = NONE;
    p->eqvPreprocessing = eqvPreprocessing;
    p->clusterize = clusterize;
    p->ovnum = vnum;
    p->min_clu_perc = min_cl_perc;
    
    p->tdif = xmalloc((vnum + 1) * sizeof(uint));
    p->tnlt = xmalloc((vnum + 1) * sizeof(uint));
    p->tins = xmalloc((vnum + 1) * sizeof(uint));
    p->clst = (flexarray_t **)xcalloc(1, 2*(vnum + 1)*sizeof(flexarray_t *));
    p->elst = (flexarray_t **)xcalloc(1, (vnum + 1)*sizeof(flexarray_t *));
    p->pri = (uint **)xcalloc(1, lits * sizeof(uint *));
    p->clusters = NULL;
    p->depvars = initFlexArray(16);
    p->elimvars = initFlexArray(16);

    // everything else, except this block, can be freed
    // after the preprocessing is over
    // --- BEGIN BLOCK ---
    // there can be at most clnum/2 binary equivalence clauses
    // at any point during the preprocessing
    p->eqs = (cla_t **)xcalloc(1, (clnum >> 1) * sizeof(cla_t *));
    // p->eqs can't be deleted because it's a container for 
    // equivalence clauses that are freed at the end
    p->model = xmalloc((vnum + 1) * sizeof(uint));
    p->taut_stack = initStack(16);
    // --- END BLOCK ---

    p->impq = initBcpq(vnum);
    p->eq_graph = grph_new_al_grph(vnum);
    p->ocnf = clst;
    p->ecnf = initFlexArray(16);
    p->eqs_ptr = 0;

    for (i = 0; i < lits; i++) {
        // init occurrence counters
        p->pri[i] = (uint *)xcalloc(1, 4 * sizeof(uint));
        // init CNF clause lists
        p->clst[i] = initFlexArray(4);
    }
    for (i = 0; i < vnum + 1; i++) {
        // initially none of the variables are mapped
        p->model[i] = UINT_MAX;
        // init equivalence list
        p->elst[i] = initFlexArray(4);
    }
    return p;
}

void partiallyFreePreproc(preproc_t *p) {
    uint i, lits = (p->ovnum + 1) << 1;
    if (p->clusters) {
        uint *a = getFlexArray(p->clusters);
        uint sz = getFlexArraySize(p->clusters);
        for (i = 0; i < sz; i++) {
            flexarray_t *f = (flexarray_t *)a[i];
            freeFlexArray(f);
            a[i] = 0;
        }
        freeFlexArray(p->clusters);
        p->clusters = NULL;
    }
    freeFlexArray(p->depvars);
    p->depvars = NULL;
    freeFlexArray(p->elimvars);
    p->elimvars = NULL;
    free(p->tdif);
    p->tdif = NULL;
    free(p->tnlt);
    p->tnlt = NULL;
    free(p->tins);
    p->tins = NULL;
    freeBcpq(p->impq);
    p->impq = NULL;
    free_al_grph(p->eq_graph);
    p->eq_graph = NULL;
#ifdef DEBUGX
    uint *a = getFlexArray(p->ocnf);
    uint sz = getFlexArraySize(p->ocnf);
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        assert(c);
        assert(getClSize(c) > 1);
    }
#endif
    for (i = 0; i < p->eqs_ptr; i++) {
        cla_t *c = p->eqs[i];
        if (getClSize(c) == 0) {
            free(c);
            p->eqs[i] = NULL;
        }
    }
    for (i = 0; i < lits; i++) {
        freeFlexArray(p->clst[i]);
        free(p->pri[i]);
    }
    for (i = 0; i < (p->ovnum + 1); i++) {
        freeFlexArray(p->elst[i]);
    }
    free(p->ecnf);
    p->ecnf = NULL;
    free(p->pri);
    p->pri = NULL;
    free(p->clst);
    p->clst = NULL;
    free(p->elst);
    p->elst = NULL;
}

void freePreproc(preproc_t *p) {
    uint i;
    free(p->model);
    for (i = 0; i < p->eqs_ptr; i++) {
        if (p->eqs[i]) {
            free(p->eqs[i]);
        }
    }
    free(p->eqs);
    freeClLst(p->ocnf);
    stackFree(p->taut_stack);
    freeFlexArray(p->ocnf);
    free(p);
}

#ifdef DEBUG
void printCNF(preproc_t *p, char *ofile) {
    FILE *fp = NULL;
    uint i, sz, *a;
    char *msg = xmalloc(4096);
    a = getFlexArray(p->ocnf);
    sz = getFlexArraySize(p->ocnf);

    // create output file if needed
    if (ofile) {
        fp = fopen(ofile, "w+");
        if (!fp) {
            error("Output file creation error!");
        }
    }
    // write p line
    sprintf(msg, "p cnf %u %u\n", p->ovnum, sz);
    writeStream(msg, fp);

    // print tautologies
    while (!stackEmpty(p->taut_stack)) {
        cla_t *x = (cla_t *)stackPop(p->taut_stack);
        if (!x) continue;
        assert(isEquiv(x));
        printCNFClause(msg, x, fp);
    }

    for (i = 0; i < p->eqs_ptr; i++) {
        cla_t *c = p->eqs[i];
        if (getClSize(c) == 0 || isTautology(c)) {
            continue;
        }
        printCNFClause(msg, c, fp);
    }

    // write all clauses in CNF format 
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        printCNFClause(msg, c, fp);
    }
    // clean-up 
    free(msg);
    if (fp) fclose(fp);
}
#endif

void trimClLst(flexarray_t *lst) {
    uint sz = getFlexArraySize(lst);
    uint *a = getFlexArray(lst);
    if (sz > 0) {
        uint i = sz - 1;
        do {
            cla_t *c = (cla_t *)a[i];
            uint csz = getClSize(c);
            if (csz) break;
            free(c);
        } while (i--);
        lst->idx = i + 1;
    }
}

void purgeZeroSized(flexarray_t *lst) {
    int i, lasti = -1;
    uint *a = getFlexArray(lst);
    uint sz = getFlexArraySize(lst);
    for (i = 0; (uint)i < sz; i++) {
        cla_t *x = (cla_t *)a[i];
        uint xsz = getClSize(x);
        if (xsz > 0 && lasti > -1) {
            assert(lasti != i);
            a[lasti++] = a[i];
        } else if (lasti < 0 && xsz == 0) {
            lasti = i;
        }
    }
    lst->idx = (uint)_max(lasti, 0);
}

void update_model(preproc_t *p, uint v, uint m) {
    assert(p->model[v] > 1);
    p->model[v] = m;
}
        
int appendInt(flexarray_t *cls, int x, flexarray_t *f) {
    assert(f != NULL);
    if (x != 0) {
        appendToFlexArray(f, formVar(x));
        return 0;
    } else {
        uint *a = getFlexArray(f);
        uint sz = getFlexArraySize(f);
        cla_t *ncl;
        if (sz == 0) {
            warrn("Empty clause found - problem UNSAT!");
            return 0;
        }
        ncl = getNewClause(a, sz);
        appendToFlexArray(cls, (uint)ncl);
        resetFlexArray(f);
        return 1;
    }
}

flexarray_t *parse(char *name, uint *vnum, uint *clnum) {
    long tmp_fsize;
    int fsize;
#ifdef USE_MMAP
    int fp = open(name, O_RDONLY);
    if (fp <= 0) {
        error("Could not open input file!");
    }
#else
    FILE *fp = fopen(name, "rb");
#endif
    uint clcnt = 0;
    flexarray_t *cls = initFlexArray(1024);
    flexarray_t *tcl = initFlexArray(32);
    char *buf, *ptr, *left, *tptr;

    if (!fp) error("Input file open error!");
#ifdef USE_MMAP
    tmp_fsize = lseek(fp, 0L, SEEK_END);
    if (tmp_fsize > INT_MAX) {
        error("File too large!");
    } else {
        fsize = (int)tmp_fsize;
    }
    ptr = buf = xmmap(0, fsize, PROT_READ, MAP_PRIVATE, fp, 0);
#else
    if (fseek(fp, 0L, SEEK_END)) {
        error("File handling error!");
    }
    tmp_fsize = ftell(fp);
    if (tmp_fsize > INT_MAX) {
        error("File too large!");
    } else {
        fsize = (int)tmp_fsize;
    }
    rewind(fp);
    ptr = buf = xmalloc((size_t)fsize);
    (void)fread(buf, 1, (size_t)fsize, fp);
#endif
    
    left = ptr + fsize;
    while (ptr < left) {
        if (isspace(*ptr)) {
            ptr++;
        } else if (isdigit(*ptr) || *ptr == '-') {
            int x = (int)strtol(ptr, &tptr, 10);
            if (ptr != tptr) { // valid int read
                if (*vnum && (uint)abs(x) > *vnum) {
                    char *msg = xmalloc(128);
                    sprintf(msg, "Invalid literal: %d", x);
                    error(msg);
                } 
                clcnt += appendInt(cls, x, tcl);
            }
            ptr = tptr;
        } else if (*ptr == 'c') {
            ptr = memchr(ptr, '\n', MAX_LINE_LENGTH);
        } else if (*ptr == 'p') {
            ptr++;
            (void)strtol(ptr, &ptr, 36); // skip over cnf token
            *vnum = (uint)strtol(ptr, &tptr, 10);
            assert(ptr != tptr);
            ptr = tptr;
            *clnum = (uint)strtol(ptr, &tptr, 10);
            assert(ptr != tptr);
            ptr = tptr;
        } else {
            char *msg = xmalloc(512);
            sprintf(msg, "Invalid line: %s", ptr);
            error(msg);
        }
    } 
    
    if (*vnum == 0 || *clnum == 0) {
        error("P line not found in CNF file!");
    } else if (*clnum != clcnt) {
        error("Invalid number of clauses in P line!");
    }
    
#ifdef USE_MMAP
    munmap(buf, fsize);
    close(fp);
#else
    free(buf);
    fclose(fp);
#endif
    freeFlexArray(tcl);
    return cls;
}

int clcmp(const void *ca, const void *cb) {
    cla_t *a = (cla_t *)(*((uint *)ca));
    cla_t *b = (cla_t *)(*((uint *)cb));
    uint sza = getClSize(a), szb = getClSize(b);
    uint xa, xb;
    assert(!isEquiv(a));
    assert(!isEquiv(b));

    if (sza != szb) {
        if (sza == 0) {
            return 1;
        } else if (szb == 0) {
            return -1;
        } else {
            return (sza < szb) ? -1 : 1;
        }
    } else if (sza == 0) return 0;
    
    assert(sza == szb);
    assert(sza > 1);

    sza--;
    do {
        xa = a[sza] >> 1; 
        xb = b[sza--] >> 1;
        if (xa != xb) break;
    } while (sza);
    if (xa != xb) return (xa < xb) ? -1 : 1;

    szb--;
    do {
        xa = a[szb];
        xb = b[szb--];
        if (xa != xb) break;
    } while (szb); 
    return (xa != xb) ? (xa < xb) ? -1 : 1 : 0;
}

void addcl_cnf_aux(preproc_t *p, cla_t *c) {
    uint i, sz = getClSize(c);
    assert(sz > 1);
    assert(!isEquiv(c));
    if (sz == 2) {
        if (impqPut(p->impq, c[1]) == BCP_CONFLICT) {
            p->sat_status = FALSE;
        }
        setClSize(c, 0);
    } else {
        for (i = 1; i < sz; i++) {
            uint lit = c[i];
            assert(lit > 1 && lit < (p->ovnum + 1) << 1);
            assert(bfsearchFlexArray(p->clst[lit], (uint)c, 1) < 0);
            appendToFlexArray(p->clst[lit], (uint)c);
            p->pri[lit][PRI_EXACT]++;
        }
    }
}

void addcl_coe_aux(preproc_t *p, cla_t *c) {
    uint i, idx = p->eqs_ptr, sz = getClSize(c);
    assert(sz > 3);
    assert(isEquiv(c));
    for (i = 1; i < sz; i++) {
        uint v = c[i];
        assert(bfsearchFlexArray(p->elst[v], idx, 1) < 0);
        appendToFlexArray(p->elst[v], idx);
    }
    // add it to eqs list
    assert(p->eqs[idx] == NULL);
    p->eqs[p->eqs_ptr++] = c;
}

void populate_aux(preproc_t *p, flexarray_t *formula, flexarray_t **tbl) {
    uint *a, i, sz;
    a = getFlexArray(formula);
    sz = getFlexArraySize(formula);
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        assert(!isEquiv(c));
        if (getClSize(c) == 0) {
            continue;
        }
        addcl_cnf_aux(p, c);
    }
}

// delete one clause from the list of literal/variable idx
void delone_aux(flexarray_t **tbl, uint idx, uint dat) {
    flexarray_t *f = tbl[idx];
    int i = bfsearchFlexArray(f, dat, 1);
#ifdef DEBUG
    uint sz = getFlexArraySize(f);
#endif
    assert(i >= 0);
    assert(getFlexArray(f)[i] == dat);
    delFlexElem(f, (uint)i);
    assert(bfsearchFlexArray(f, dat, 1) == -1);
#ifdef DEBUG
    assert(getFlexArraySize(f) == sz - 1);
#endif
}

void delcl_cnf_aux(preproc_t *p, cla_t *c) {
    uint i, sz = getClSize(c);
    assert(!isEquiv(c));
    assert(sz > 1);
    for (i = 1; i < sz; i++) {
        delone_aux(p->clst, c[i], (uint)c);
    }
    setClSize(c, 0);
    statsUpdate(p->stats, PP_EVNT_NUM_ELIM_CLS, 0);
}

void delcl_coe_aux(preproc_t *p, cla_t *c, uint cidx) {
    uint i, sz = getClSize(c);
    assert(isEquiv(c));
    assert(sz > 1);
    for (i = 1; i < sz; i++) {
        delone_aux(p->elst, c[i], cidx);
    }
    setClSize(c, 0);
}
        
void delset_cnf_aux(preproc_t *p, flexarray_t **tbl, uint lit) {
    uint sz, *a;
    int i;
    a = getFlexArray(tbl[lit]);
    sz = getFlexArraySize(tbl[lit]);
    // for each clause
    for (i = (int)sz - 1; i >= 0; i--) {
        cla_t *c = (cla_t *)a[i];
        assert(!isEquiv(c));
        assert(getClSize(c) > 1);
        delcl_cnf_aux(p, c);
    }
    resetFlexArray(tbl[lit]);
}

// removes all occurrences of lit in clause lists
void dellit_aux(preproc_t *p, uint lit, cla_t **eqs) {
    uint *a, sz, idx, i;
    flexarray_t *clf;
    idx = (eqs) ? getVar(lit) : lit;
    clf = (eqs) ? p->elst[idx] : p->clst[idx];
    a = getFlexArray(clf);
    sz = getFlexArraySize(clf);
    // for each clause (or clause index for equivalencies)
    if (eqs) {
        for (i = 0; i < sz; i++) {
            uint sgn, csz, dat = a[i];
            cla_t *c = eqs[dat];
            sgn = getEquivSgn(c) ^ (LSB(lit) ^ 1);
            assert(isEquiv(c));
            // removing of negative literal from the equivalency
            // changes its sign (1 means negative for equivalencies
            // and positive for literals)
            setEquivSgn(c, sgn);
            assert(getClSize(c) > 2);
            assert(checkClSort(c));
            assert(!isTautology(c));
            delLit(c, idx);
            csz = getClSize(c);
            if (isUnitSz(csz)) { // unit clause
                elimUnitCl(p, c, (int)dat);
            } 
        }
    } else {
        for (i = 0; i < sz; i++) {
            cla_t *c = (cla_t *)a[i];
            assert(!isEquiv(c));
            assert(getClSize(c) > 2);
            assert(checkClSort(c));
            delLit(c, idx);
            if (isUnitCl(c)) {
                elimUnitCl(p, c, -1);
            }
        }
    }
    resetFlexArray(clf);
}

// replace l1 with l2
void repset_aux(preproc_t *p, uint l1, uint l2) {
    uint *a, i;
    a = getFlexArray(p->clst[l1]);
    for (i = 0; i < getFlexArraySize(p->clst[l1]); i++) {
        cla_t *c = (cla_t *)a[i];
        int k = findVar(c, l2);
        assert(checkClSort(c));
        assert(!isEquiv(c));
        assert(getClSize(c) > 1);
        if (k > 0) {
            uint xlit = c[k];
            // literal is deleted right here for two reasons:
            // 1) we avoid brute force search for the clause
            // in the list of l1's clauses
            // 2) we don't need to update the i index after
            // deleting the clause
            delLit(c, l1);
            // if double literal is found, delete 
            // literal which needs to be replaced
            if (xlit == l2) {
                if (isUnitCl(c)) {
                    elimUnitCl(p, c, -1);
                }
            } else { // tautology
                assert(xlit == (l2 ^ 1));
                delcl_cnf_aux(p, c);
            }
        } else {
            replaceLit(c, l1, l2);
            cmp_qsort(&(c[1]), getClSize(c) - 1, 4, ucmp);
            assert(bfsearchFlexArray(p->clst[l2], (uint)c, 1) < 0);
            appendToFlexArray(p->clst[l2], (uint)c);
            assert(checkClSort(c));
        }
    }
    resetFlexArray(p->clst[l1]);
}

void elimUnitCl(preproc_t *p, cla_t *c, int eqv_idx) {
    uint lit, dat, idx;
    flexarray_t **lst;
    assert(isUnitCl(c));
    if (isEquiv(c)) {
        assert(eqv_idx >= 0);
        assert(getEquivSgn(c) < 2);
        lit = (c[1] << 1) | getEquivSgn(c);
        idx = c[1];
        lst = p->elst;
        dat = (uint)eqv_idx;
    } else {
        assert(eqv_idx < 0);
        idx = lit = c[1];
        lst = p->clst;
        dat = (uint)c;
        statsUpdate(p->stats, PP_EVNT_NUM_ELIM_CLS, 0);
    }
    delone_aux(lst, idx, dat);
    setClSize(c, 0);
    if (impqPut(p->impq, lit) == BCP_CONFLICT) {
        p->sat_status = FALSE;
    } 
}

// eqv = 1 if v1 <=> v2
// replace v1 with v2
void repset_eqv_aux(preproc_t *p, uint v1, uint v2, uint eqv) {
    uint *a, i, sz;
    a = getFlexArray(p->elst[v1]);
    sz = occurInCOE(p, v1);
    for (i = 0; i < sz; i++) {
        uint cli = a[i];
        cla_t *c = p->eqs[cli];
        int k = findLit(c, v2);
        assert(checkClSort(c));
        assert(isEquiv(c));
        assert(getClSize(c) > 2);
        assert(getEquivSgn(c) < 2);
        setEquivSgn(c, getEquivSgn(c) ^ (eqv ^ 1));
        if (k > 0) {
            assert(c[k] == v2); // check search result
            assert(checkClSort(c));
            delLitAt(c, (uint)k); // del v2
            delone_aux(p->elst, v2, cli);
            delLit(c, v1);
            if (isUnitCl(c)) {
                elimUnitCl(p, c, (int)cli);
            }
                    
            // binary equivalences must not be detected and
            // eliminated here, as repset_eqv_aux is called
            // from propagate_bieq(), and we need to finish
            // propagation of root variable and unmark the root
            // before eliminating those binary equivalences
        } else {
            assert(v1 != c[getClSize(c)]);
            replaceLit(c, v1, v2);
            cmp_qsort(&(c[1]), getClSize(c) - 1, 4, ucmp);
            assert(bfsearchFlexArray(p->elst[v2], cli, 1) < 0);
            appendToFlexArray(p->elst[v2], cli);
            assert(checkClSort(c));
        }
    }    
    resetFlexArray(p->elst[v1]);
}

char impqPut(bcpq_t *iq, uint lit) {
    uint v = getVar(lit);
    uint res, sgn = getPhase(lit) + 1;
    // POS (10) if 1, NEG (01) if 0
    res = writeTable(iq->t, v, sgn);
    if (res == WRITE_SUCCESS) {
        iq->a[iq->widx++] = lit;
        return BCP_OK;
    } else if (res == sgn) {
        return BCP_OK;
    }
    // conflict
    bcpqReset(iq);
    return BCP_CONFLICT;
}

uint impqGet(bcpq_t *iq) {
    assert(!bcpqEmpty(iq));
    return iq->a[iq->ridx++];
}

uint bineq_unroll(preproc_t *p, uint v, uint lit) {
	uint model;
    assert(v > 0 && v <= p->ovnum);
    model = p->model[v];
    assert(model > 1);
    if (model == UINT_MAX) {
        return lit;
    } else {
        uint replit, mdv = getVar(model);
        assert(mdv > 0 && mdv <= p->ovnum);
        replit = bineq_unroll(p, mdv, model);
        return replit ^ (LSB(lit) ^ 1);
    }
}

// propagates unit variables. at the end sorts clauses, but doesn't
// free zero-sized clauses.
char bcp(preproc_t *p) {
    char change = FALSE;
    // there must be no binary equivalences waiting to be propagated
    assert(!grph_any_eqs(p->eq_graph));
    while (!bcpqEmpty(p->impq)) { 
        uint lit = impqGet(p->impq);
        uint v = getVar(lit);
        uint model = bineq_unroll(p, v, lit);
        if (model != lit) {
            char res = impqPut(p->impq, model);
            if (res == BCP_CONFLICT) {
                p->sat_status = FALSE;
                break;
            }
            continue;
        }
        statsUpdate(p->stats, PP_EVNT_UNITV, 0);
        assert(v > 0 && v <= p->ovnum);
        appendToFlexArray(p->elimvars, v);
        change = TRUE;
        // eliminate all satisfied clauses
        delset_cnf_aux(p, p->clst, lit);
        // eliminate lit from all unsat clauses, add new unit lits
        dellit_aux(p, lit ^ 1, NULL);
        // eliminate var from all equivalencies
        dellit_aux(p, lit, p->eqs);
        // update model
        update_model(p, v, LSB(lit));
        if (p->sat_status != NONE) break;
    }
    return change;
}

char find_bieq(preproc_t *p) {
    uint i;
    char bin_found = FALSE;
    for (i = 0; i < p->eqs_ptr; i++) {
        cla_t *c = p->eqs[i];
        uint csz = getClSize(c);
        assert(isEquiv(c));
        assert(!isBinSz(csz) || !c[csz]);
        if (csz == 0 || c[csz]) continue;
        if (isBinSz(csz)) {
            bin_found = TRUE;
            assert(getEquivSgn(c) < 2);
            grph_eqv_insert(p->eq_graph, c[1], c[2], getEquivSgn(c));
            delcl_coe_aux(p, c, i);
        }
    }
    return bin_found;
}

// propagates binary equivalencies, and sorts clauses at the
// end if any change is made. zero-sized clauses are not freed!
char propagate_bieq(preproc_t *p) {
    char change = FALSE;
    // there must not be any implied variables waiting to be
    // propagated
    assert(bcpqEmpty(p->impq));
    do {
        while (grph_any_eqs(p->eq_graph)) {
            uint *a, j = 0, rootv;
            a = grph_alt_dfs(p->eq_graph);
            if (!a) { // invalid equivalency detected
                p->sat_status = FALSE;
                return TRUE;
            }
            if (!a[j++]) { // skip visited and empty nodes
                continue;
            }
            rootv = getVar(a[0]);
            assert(rootv > 0 && rootv <= p->ovnum);
            assert(LSB(a[0]) == 1);
            change = TRUE;
            while (a[j]) {
                uint v = getVar(a[j]);
                appendToFlexArray(p->elimvars, v);
                assert(v <= p->ovnum);
                // replace literal in CNF clauses
                repset_aux(p, a[j], a[0]);
                // replace var of l1 with var of l2 in all equivalencies
                repset_eqv_aux(p, v, rootv, LSB(a[j]));
                // update model:
                // LSB(x) = 1 if getVar(x) <=> i, 1 otherwise
                update_model(p, v, makeLit(rootv, LSB(a[j])));
                // invert the literal for second pass
                a[j] ^= 1;
                j++;
            }
            a[0] ^= 1;
            j = 1;
            while (a[j]) {
                repset_aux(p, a[j], a[0]);
                j++;
            }
            grph_reset_root(p->eq_graph, rootv);
            if (p->sat_status != NONE) return TRUE;
        }
        // find_bieq might find new binary equivalences. but,
        // some variables participating in binary equivalences
        // might already be implied, so always before detecting
        // new binary equivalences, there must be no implied
        // variables.
        (void)bcp(p);
        assert(!grph_any_eqs(p->eq_graph));
    } while (find_bieq(p));
    return change;
}


#define CUTOFF 4

// one-step resolution, at the end sorts cnf clauses and
// frees zero-sized clauses
char onestep_resolve(preproc_t *p) {
    char loc_changes, changes = FALSE, deleted = FALSE;
    do {
        uint *a, sz, i, lim;
        cla_t *c;
        a = getFlexArray(p->ocnf);
        sz = getFlexArraySize(p->ocnf);
        loc_changes = FALSE;
        deleted = FALSE;
        lim = (sz) ? sz - 1 : 0;
        // for each clause in CNF
        for (i = 0; i < lim; i++) {
            uint j, dif = 0, cs, idx = 0;
            cla_t *nxtc;
            c = (cla_t *)a[i];
            cs = getClSize(c);
            assert(!isEquiv(c));
            if (cs == 0) {
                i++;
                break;
            }
            assert(cs > 1);
            nxtc = (cla_t *)a[i + 1];
            if (cs != getClSize(nxtc)) {
               continue;
            }
            for (j = 1; j < cs; j++) { 
                uint ldif = c[j] ^ nxtc[j];
                dif += ldif;
                if (dif > 1) {
                    break;
                }
                if (ldif == 1) {
                    assert(idx == 0);
                    idx = j;
                }
            }
            if (dif == 1 && cs < CUTOFF) {
                if (isUnitCl(nxtc)) {
                    p->sat_status = FALSE;
                    return TRUE;
                }
                // delete literal at spot idx
                delone_aux(p->clst, nxtc[idx], (uint)nxtc);
                // delete the literal in the clause
                delLitAt(nxtc, idx);
                if (isUnitCl(nxtc)) {
                    // and the whole clause if it became 
                    // a unit clause
                    elimUnitCl(p, nxtc, -1);
                }
                // skip nxtc in the next iteration
                i++;
                delcl_cnf_aux(p, c);
                loc_changes = TRUE;
            } 
            if (dif == 0) {
                delcl_cnf_aux(p, c);
                deleted = TRUE;
            }
        }

        if (loc_changes) {
            changes = TRUE;
            (void)bcp(p);
            if (p->sat_status != NONE) return TRUE;
            cmp_qsort(a, i, 128, clcmp);
        }
        trimClLst(p->ocnf);
    } while (loc_changes);
    if (deleted) {
        purgeZeroSized(p->ocnf);
    }
    return changes;
}// */

// finds equivalencies, and sorts clauses at the end if any
// change is made, but doesn't free zero-sized clauses.
char eq_find(preproc_t *p) {
    uint *a, i = 0, sz;
    char change = FALSE;
    a = getFlexArray(p->ocnf);
    sz = getFlexArraySize(p->ocnf);
    for (; i < sz - 1; i++) {
        cla_t *c = (cla_t *)a[i], *xc;
        uint xcsz, csz = getClSize(c);
		uint esz, j, c_xor = 0, xc_xor = 0, lasti;
		uint c_sgn = FALSE, xc_sgn = FALSE;
        if (csz == 0) {
#ifdef DEBUG
            for (; i < sz; i++) {
                cla_t *x = (cla_t *)a[i];
                assert(getClSize(x) == 0);
            }
#endif
            break;
        }
        esz = 1 << (csz - 2);
        // index of the last possible clause which
        // should be a part of equivalence
        lasti = i + esz - 1;
        // as eq_find is after bcp call (which eliminates
        // unit clauses and sorts all clauses at the end),
        // there should be no zero-sized or unit clauses
        assert(csz > 2);
        assert(!isEquiv(c));
        if (lasti >= sz) {
            break;
        }
        xc = (cla_t *)a[lasti];
        xcsz = getClSize(xc);
        assert(!isEquiv(xc));
        if (xcsz == 0) {
            break;
        } else if (csz != xcsz) {
            continue;
        }
        assert(xcsz > 2);
        // compare variables and signs of the first and
        // last clause in the domain
        for (j = 1; j < csz; j++) {
            c_xor ^= c[j];
            xc_xor ^= xc[j];
            if ((c[j] ^ xc[j]) > 1) {
                goto repeat_outer_loop;
            }
        }
        c_sgn = LSB(c_xor);
        xc_sgn = LSB(xc_xor);
        if (c_sgn != xc_sgn) {
repeat_outer_loop:
            continue;
        }
        // at this point we know that the first and last
        // clauses have the same variables, and the same sign,
        // here we check signs of all other clauses in the domain
        for (j = i + 1; j < lasti; j++) {
            uint k;
            xc_xor = 0;
            xc = (cla_t *)a[j];
            assert(getClSize(xc) > 2);
            assert(getClSize(xc) == csz);
            assert(!isEquiv(xc));
            for (k = 1; k < csz; k++) {
                xc_xor ^= xc[k];
                assert((xc[k] ^ c[k]) <= 1);
            }
            xc_sgn = LSB(xc_xor);
            if (c_sgn != xc_sgn) {
                // the first and the last clause in the domain
                // have the same sign and variables. if one of
                // the clauses in the domain has different sign,
                // we can continue the search from the next one
                // in the domain
                i = j + 1;
                break;
            }
        }
        if (c_sgn != xc_sgn) {
            continue;
        }
        change = TRUE;
#ifdef DEBUG
        {
            // check that all the clauses are actually different
            uint dbgi, dbgj, dbgk;
            for (dbgi = i; dbgi < lasti; dbgi++) {
                for (dbgj = dbgi + 1; dbgj <= lasti; dbgj++) {
                    uint dif = 0;
                    cla_t *dbgc1 = (cla_t *)a[dbgi];
                    cla_t *dbgc2 = (cla_t *)a[dbgj];
                    for (dbgk = 1; dbgk < csz; dbgk++) {
                        dif += dbgc1[dbgk] ^ dbgc2[dbgk];
                    }
                    assert(dif > 0);
                }
            }
        }
#endif
        // if it's a binary equivalence, add it to bineq graph
        if (csz == 3) {
            grph_eqv_insert(p->eq_graph, getVar(c[1]), getVar(c[2]), c_sgn);
        } else { 
            // otherwise we create an equivalency clause...
            cla_t *eqc = getNewEqvCla(c, c_sgn);
            assert(csz > 3);
            assert(isEquiv(eqc));
            assert(checkClSort(eqc));
            assert(getClSize(eqc) > 2);
            // add it to aux_tbl
            addcl_coe_aux(p, eqc); 
        }
        // at this point, we have detected an equivalence, so
        // so we delete CNF clauses forming an equivalence
        for (j = i; j <= lasti; j++) {
            // del clause from aux_tbl
            delcl_cnf_aux(p, (cla_t *)a[j]);
        }
        i = lasti;
    }
    return change;
}

void elimTautology(preproc_t *p, cla_t *c, uint cl_idx) {
    uint k, csz;
    assert(isEquiv(c));
    assert(!isTautology(c));
    // mark it as tautology
    markTautology(c);
    csz = getClSize(c);
    assert(!c[csz]);
    // delete the clause from aux_tbl, this is
    // necessary so that other tautologies can
    // be detected more easily
    for (k = 1; k < csz; k++) {
        delone_aux(p->elst, c[k], cl_idx);
    }    
    // add it to tautology stack
    stackPush(p->taut_stack, (uint)c);
}

inline uint occurInCOE(preproc_t *p, uint v) {
    return getFlexArraySize(p->elst[v]);
}

inline uint occurInCNF(preproc_t *p, uint v) {
    uint l = v << 1;
    return getFlexArraySize(p->clst[l]) + 
        getFlexArraySize(p->clst[l+1]);
}

// returns TRUE if any binary or unit eqv clauses were generated.
// both types of clauses are immediately deleted. bin equivalence
// relation is entered to graph, while unit eqv literals are 
// added to implication graph.
char eq_reason(preproc_t *p) {
    uint i;
    char unit_or_bin_generated = FALSE;
    // for all equivalencies
    for (i = 0; i < p->eqs_ptr; i++) {
        uint j, csz;
        // variable with max reduction
        cla_t *c = p->eqs[i];
        csz = getClSize(c);
        assert(i < p->eqs_ptr);
        assert(isEquiv(c));
        // skip zero-sized clauses, clauses with
        // determined dependant variable and tautologies
        if (csz == 0 || c[csz])  continue; 

        // for each variable in the clause
        for (j = 1; j < csz; j++) {
            uint k, v = c[j], *cfa, cfsz;
            assert(v > 0 && v <= p->ovnum);
            cfa = getFlexArray(p->elst[v]);
            cfsz = occurInCOE(p, v); 
            for (k = 0; k < cfsz; k++) { 
                char st;
                if (i == cfa[k]) continue;
                st = eq_reduce(p, i, cfa[k], TRUE);
                switch (st) {
                    case EQR_UNIT_OR_BIN: // unit/bin cl generated
                        unit_or_bin_generated = TRUE;
                    case EQR_REGULAR: // regular reduction
                        break;
                    case EQR_EMPTY: // empty clause generated
                        return TRUE;
                }
                k--; cfsz--;
            }
        }
    }
    return unit_or_bin_generated;
}

// reduce c2 with c1
char eq_reduce(preproc_t *p, uint c1i, uint c2i, char onlyIfShorter) {
    uint j, c2sz;
    uint *dif = p->tdif;
    uint *nlt = p->tnlt;
    uint *ins = p->tins;
    uint sgn_exor;
    cla_t *c1, *c2;
    c1 = p->eqs[c1i];
    c2 = p->eqs[c2i];
    // as unit and bin clauses are immediately eliminated,
    // the size of both c1 and c2 must be > 2
    assert(getClSize(c1) > 2);
    assert(getClSize(c2) > 2);
    assert(!isTautology(c1));
    assert(!isTautology(c2));
    assert(!c1[getClSize(c1)]);
    assert(c1i != c2i);
    c2sz = getClSize(c2);
    assert(getEquivSgn(c1) < 2);
    assert(getEquivSgn(c2) < 2);
    sgn_exor = !(getEquivSgn(c1) ^ getEquivSgn(c2));
    clDif(c1, c2, dif, nlt, ins);
    // if reduced clause is either:
    // empty (redundant or invalid equivalence), unit or bin,
    // remove it from aux_tbl
    if (dif[0] < 4) {
        delcl_coe_aux(p, c2, c2i);
    }
    if (dif[0] == 1) { // check for invalid equivalencies
        // if we have two equal equivalencies with
        // the opposite signs, that's unsatisfiable
        if (!sgn_exor) {       
            p->sat_status = FALSE;
            return EQR_EMPTY;
        }
    } else if (dif[0] == 2) { // unit clause detected
        uint lit = makeLit(dif[1], sgn_exor);
        char res = impqPut(p->impq, lit);
        if (res == BCP_CONFLICT) {
            p->sat_status = FALSE;
            return EQR_EMPTY;
        }
        return EQR_UNIT_OR_BIN;
    } else if (dif[0] == 3) { // bin equivalence detected
        grph_eqv_insert(p->eq_graph, dif[1], dif[2], sgn_exor);
        return EQR_UNIT_OR_BIN;
    } else if (c2sz > dif[0] || !onlyIfShorter) {
        // create new clause
        cla_t *x = formEqvCla(dif, sgn_exor);
        // set its dependant variable
        x[getClSize(x)] = c2[c2sz];
        assert(c2[c2sz] == 0 || findLit(x,c2[c2sz]) > 0);
        assert(checkClSort(x));
        assert(getClSize(x) > 1);
        // eliminate removed vars from aux_tbl
        for (j = 1; j < ins[0]; j++) {
#ifdef DEBUG
            uint nume = p->elst[ins[j]]->idx;
            assert(bfsearchFlexArray(p->elst[ins[j]], c2i, 1) >= 0);
#endif
            delone_aux(p->elst, ins[j], c2i);
#ifdef DEBUG
            assert(p->elst[ins[j]]->idx == nume - 1);
            assert(bfsearchFlexArray(p->elst[ins[j]], c2i, 1) == -1);
#endif
        }
        // add new vars to aux_tbl
        for (j = 1; j < nlt[0]; j++) {
            assert(bfsearchFlexArray(p->elst[nlt[j]], c2i, 1) < 0);
            appendToFlexArray(p->elst[nlt[j]], c2i);
        }
        // free old clause
        free(c2);
        // and add new one
        p->eqs[c2i] = x;
    } 
    return EQR_REGULAR;
}

// returns TRUE if any binary or unit eqv clauses were generated.
// both types of clauses are immediately deleted. bin equivalence
// relation is entered to graph, while unit eqv literals are 
// added to implication graph.
char eq_find_depvars(preproc_t *p) {
    uint i;
    char unit_or_bin_generated = FALSE;
    for (i = 0; i < p->eqs_ptr; i++) {
        cla_t *c = p->eqs[i];
        uint csz = getClSize(c);
        uint j, min_cnf_vidx = 0, min_cnf = UINT_MAX;
        assert(isEquiv(c));
        // skip zero-size clauses, clauses with determined
        // dependant variable and tautologies
        if (csz == 0 || c[csz]) {
            continue;
        }
        assert(csz > 1);
        // dependant or tautology variable should be zero
        assert(!c[csz]);
        // for each variable in equivalence
        for (j = 1; j < csz; j++) {
            uint v = c[j];
            uint eqv_occur;
            assert(v > 0 && v <= p->ovnum);
            eqv_occur = occurInCOE(p, v);
            // otherwise find the least occuring var in CNF
            if (eqv_occur < min_cnf) {
                min_cnf = eqv_occur;
                min_cnf_vidx = j;
            }
        }
        // if not tautology
        if (csz > 0) {
			uint *cfa, cfsz, v;
            assert(min_cnf_vidx < getClSize(c));
            assert(min_cnf_vidx > 0);
            // do the reduction with the least occuring var
            v = c[min_cnf_vidx];
            assert(v > 0);
            assert(v <= p->ovnum);
            cfa = getFlexArray(p->elst[v]);
            cfsz = occurInCOE(p, v);
            for (j = 0; j < cfsz; j++) {
                char st;
                if (i == cfa[j]) continue;
                st = eq_reduce(p, i, cfa[j], FALSE);
                switch (st) {
                    case EQR_UNIT_OR_BIN: // unit/bin cl generated
                        unit_or_bin_generated = TRUE;
                    case EQR_REGULAR: // regular reduction
                        break;
                    case EQR_EMPTY: // empty clause found
                        return TRUE;
                }
                j--; cfsz--;
                assert(occurInCOE(p, v) == cfsz);
            }
            // if unit_or_bin is generated, we can't compute 
            // dependant or tautology variables for the following 
            // reason: dep or taut var might be added as a bin
            // equivalence to the equivalence graph. later some
            // other variables might be replaced by taut/dep var,
            // in which case it wouldn't be a taut/dep var anymore.
            // to prevent that case, we just postpone computing
            // of dep/taut vars in that case.
            if (unit_or_bin_generated) continue;

            // at this point we know there can be only one equivalence
            // clause with variable v, so check for tautologies
            if (occurInCNF(p, v) == 0) {
                assert(occurInCOE(p, v) == 1);
                elimTautology(p, c, i);
                assert(occurInCOE(p, v) == 0);
            } 
            // at this point clause c has dependant variable v,
            // and can be a potential tautology
            assert(!c[csz]);
            c[csz] = v;            
        }
    }
    stackPush(p->taut_stack, (Item)NULL);
    return unit_or_bin_generated;
}

uint resolveBinEq(preproc_t *p, uint v) {
	uint lit, sgn = FALSE;
    assert(v > 0 && v <= p->ovnum);
    lit = p->model[v];
    if (lit == UINT_MAX) { // NONE if var is not assigned
        sgn = NONE;
    } else if (lit < 2) { // value if var is assigned
        sgn = lit;
    } else { // recursive call if var is equivalent to x
        uint x = getVar(lit);
        statsUpdate(p->stats, PP_EVNT_BINEQ, 0);
        sgn = resolveBinEq(p, x);
        assert(sgn == NONE || sgn == TRUE || sgn == FALSE);
        if (sgn != NONE) {
            sgn ^= (LSB(lit) ^ 1);
            update_model(p, v, sgn);
        }
    }
    return sgn;
}

void resolveTautEqvCl(preproc_t *p, cla_t *c) {
    uint i, sz = getClSize(c), tautv;
    uint sgn = getEquivSgn(c), lsgn = FALSE;
    assert(c[sz] > 0 && c[sz] <= p->ovnum);
    assert(isTautology(c));
    assert(isEquiv(c));
    assert(sz > 3);
    tautv = c[sz];
    for (i = 1; i < sz; i++) {
#ifdef DEBUG
        char one_unassigned_found = FALSE;
#endif
        uint x = c[i];
        uint xm = p->model[x];
        // skip unassigned vars, which should be tautology
        // variables, but we can't verify that at this point
        // as the function is called after the aux_tbl is
        // partially freed, so we don't know the number of
        // equivalence clauses having variable x. at least
        // we can check that there's no such variable in CNF 
        if (xm == UINT_MAX) {
#ifdef DEBUG
            assert(!one_unassigned_found);
            assert(x == tautv);
            one_unassigned_found = TRUE;
#endif
            continue;
        } else if (xm > 1) {
            uint bv = getVar(xm);
            assert(bv > 0 && bv <= p->ovnum);
            xm = resolveBinEq(p, bv);
            assert(xm < 2);
        }
        lsgn ^= (xm ^ 1);
    }
    assert(tautv > 0 && tautv <= p->ovnum);
    // POS EQS: 1 - satisfied by even num of neg lits (sgn == 0)
    // NEG EQS: 0 - satisfied by odd  num of neg lits (sgn == 1)
    update_model(p, tautv, (lsgn != sgn));
}

void pure_lits(preproc_t *p) {
    uint i;
    for (i = 1; i <= p->ovnum; i++) {
        if (occurInCOE(p, i) == 0 && p->model[i] == UINT_MAX
                && !grph_is_marked(p->eq_graph, i)) {
            uint l = i << 1;
            uint neg = getFlexArraySize(p->clst[l]);
            uint pos = getFlexArraySize(p->clst[l + 1]);
            // TODO: add some debug code that actually checks
            // that impqPut doesn't return a conflict
            if (neg == 0 && pos > 0) {
                statsUpdate(p->stats, PP_EVNT_PURELIT, 0);
                (void)impqPut(p->impq, l + 1);    
            } else if (pos == 0 && neg > 0) {
                statsUpdate(p->stats, PP_EVNT_PURELIT, 0);
                (void)impqPut(p->impq, l);
            }
        }
    }
}

inline uint getNumClauses(preproc_t *p, uint v) {
    uint cl, l = v << 1;
    cl = p->pri[l][PRI_SIMPLIFIED];
    l++;
    cl += p->pri[l][PRI_SIMPLIFIED];
    cl += getFlexArraySize(p->elst[v]);
    return cl;
}

int clust_cmp(const void *v1, const void *v2) {
    flexarray_t *f1 = (flexarray_t *)(*((uint *)v1));
    flexarray_t *f2 = (flexarray_t *)(*((uint *)v2));
    uint *a1 = getFlexArray(f1);
    uint *a2 = getFlexArray(f2);
    uint sz1 = getFlexArraySize(f1);
    uint sz2 = getFlexArraySize(f2);
    assert(sz1 > 0);
    assert(sz2 > 0);
    // more constrained clusters first
    return (int)(a2[sz2 - 1])/sz2 - (a1[sz1 - 1])/sz1; 
}

//
void parseClusterFile(preproc_t *p, char *name) {
    long tmp_fsize;
    size_t fsize;
    FILE *fp = fopen(name, "rb");
    char *buf, *ptr, *tptr;
    uint treshold = p->ovnum * p->min_clu_perc / 100;
    flexarray_t *cluster = initFlexArray(16);
    flexarray_t *fringes = initFlexArray(16);
    
    if (!fp) error("Input file open error!");
    if (fseek(fp, 0L, SEEK_END)) {
        error("File handling error!");
    }
    tmp_fsize = ftell(fp);
    if (tmp_fsize > INT_MAX) {
        error("File too large!");
    } else {
        fsize = (size_t)tmp_fsize;
    }
    rewind(fp);
    buf = xmalloc(fsize);
    //(void)fread(buf, 1, fsize, fp);
    
    while (fgets(buf, fsize, fp) != NULL) {
        uint clauses = 0;
        ptr = buf;
        while (TRUE) {
            if (isspace(*ptr)) {
                ptr++;
            } else if (isdigit(*ptr)) {
                uint v = (uint)strtol(ptr, &tptr, 10);
                if (ptr != tptr) {// valid int read
                    if (v > p->ovnum) {
                        char *msg = xmalloc(128);
                        sprintf(msg, "Invalid variable in \
cluster file: %u", v);
                        error(msg);
                    } else if (v != 0) {
                        appendToFlexArray(cluster, v);
                        clauses += getNumClauses(p, v);
                    }
                    ptr = tptr;
                } 
            } else { 
                // whatever else is found, it's considered
                // as the end of line
                uint csz = getFlexArraySize(cluster);
                if (csz == 0) {
                    break;
                } 

                if (!p->clusters) {
                    p->clusters = initFlexArray(8);
                }
                
                if (csz < treshold) {
                    multiAppendFlex(cluster, fringes);
                    resetFlexArray(cluster);
                } else {
                    // last value is the total number of clauses
                    // related to the cluster
                    appendToFlexArray(cluster, clauses);
                    appendToFlexArray(p->clusters, (uint)cluster);
                    cluster = initFlexArray(16);
                }
                break;
            }
        }
    }

    if (getFlexArraySize(fringes)) {
        // fringes cluster always has the lowest priority,
        // this is achieved by setting the number of related
        // clauses to zero
        appendToFlexArray(fringes, 0);
        if (!p->clusters) {
            p->clusters = initFlexArray(8);
        }
        appendToFlexArray(p->clusters, (uint)fringes);
    } else {
        freeFlexArray(fringes);
    }
    freeFlexArray(cluster);

    if (p->clusters) {
        // sort clusters
        uint i, *a = getFlexArray(p->clusters);
        uint sz = getFlexArraySize(p->clusters);
        cmp_qsort(a, sz, 4, clust_cmp);
        for (i = 0; i < sz; i++) {
            // remove the number of clauses
            flexarray_t *f = (flexarray_t *)a[i];
            f->idx--;
        }
        p->stats->ppNumClusters = sz;
    }// */
    
    free(buf);
    fclose(fp);
}// */

void cluster(preproc_t *p) {
    uint *a, sz, i, nvars = p->ovnum + 1;
    clock_t timer, tmp_timer;
    char buf[128];
    // construct MCL header
    FILE *fp = fopen("header.hdr", "w+");
    if (!fp) error("Header file creation error!");
    fputs("(mclheader\n", fp);
    fputs("mcltype matrix\n", fp);
    sprintf(buf, "dimensions %ux%u\n", nvars, nvars);
    fputs(buf, fp);
    fputs(")\n", fp);
    fputs("(mclmatrix\n", fp);
    fputs("begin\n", fp);
    fclose(fp);

    // construct graph
    timer = clock();
    fp = fopen("data.msi", "w+");
    if (!fp) error("Raw data file creation error!");

    // dump CNF hyperedges
    a = getFlexArray(p->ocnf);
    sz = getFlexArraySize(p->ocnf);
    for (i = 0; i < sz; i++) {
        uint j, k, csz;
        cla_t *c = (cla_t *)a[i];
        csz = getClSize(c);
        for (j = 1; j < csz; j++) {
            uint v = getVar(c[j]);
            for (k = j + 1; k < csz; k++) {
                sprintf(buf, "%u\t%u:1 $\n", v, getVar(c[k]));
                fputs(buf, fp);
            }
        }
    }

    // dump COE hyperedges
    a = getFlexArray(p->ecnf);
    sz = getFlexArraySize(p->ecnf);
    for (i = 0; i < sz; i++) {
        uint j, k, csz;
        cla_t *c = (cla_t *)a[i];
        csz = getClSize(c);
        for (j = 1; j < csz; j++) {
            uint v = c[j];
            for (k = j + 1; k < csz; k++) {
                sprintf(buf, "%u\t%u:1 $\n", v, c[k]);
                fputs(buf, fp);
            }
        }
    }

    /* dump CNF hyperedges
    for (i = 2; i < 2*nvars; i++) {
        flexarray_t *f = p->clst[i];
        uint j, fsz = getFlexArraySize(f);
        uint *fa = getFlexArray(f);
        for (j = 0; j < fsz; j++) {
            cla_t *c = (cla_t *)fa[j];
            uint k, csz = getClSize(c);
            for (k = 1; k < csz; k++) {
                uint l, v = getVar(c[k]);
                for (l = k + 1; l < csz; l++) {
                    sprintf(buf, "%u\t%u:1 $\n", v, getVar(c[l]));
                    fputs(buf, fp);
                }
            }
        }
    }
    // dump COE hyperedges
    for (i = 1; i < nvars; i++) {
        flexarray_t *f = p->elst[i];
        uint j, fsz = getFlexArraySize(f);
        uint *fa = getFlexArray(f);
        for (j = 0; j < fsz; j++) {
            cla_t *c = p->eqs[fa[j]];
            uint k, csz = getClSize(c), cszm = csz - 1;
            for (k = 1; k < cszm; k++) {
                uint l;
                for (l = k + 1; l < csz; l++) {
                    sprintf(buf, "%u\t%u:1 $\n", c[k], c[l]);
                    fputs(buf, fp);
                }
            }
        }
    }// */
    fclose(fp);
    system("mcxassemble -raw data.msi -hdr header.hdr -re add -r add \
-o graph.tmp 2> /dev/null");
    tmp_timer = clock();
    statsUpdate(p->stats, PP_TIME_GRAPHGEN, tmp_timer - timer);

    timer = tmp_timer;
    system("./cluster graph.tmp cluster.dat");
    // read and parse cluster.dat file
    parseClusterFile(p, "cluster.dat");
    system("rm -fr data.msi graph.tmp cluster.dat header.hdr");
    tmp_timer = clock();
    statsUpdate(p->stats, PP_TIME_CLUSTER, tmp_timer - timer);
}// */

void simplify(preproc_t *p) {
    uint *a = getFlexArray(p->ocnf);
    uint i, j, sz = getFlexArraySize(p->ocnf);
    char change, first = TRUE;

    // collect clause stats and sort each clause
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        uint x = 0, csz = getClSize(c);
        assert(csz > 1);
        for (j = 1; j < _umin(csz, 3); j++) {
            p->pri[c[j]][PRI_APPROX]++;
        }
        cmp_qsort(&(c[1]), csz - 1, 4, ucmp);

        // for each literal
        for (j = 1; j < getClSize(c); j++) {
            // check for tautologies and double literals
            if (j > 1) {
				uint ex;
                assert(x <= c[j]);
                // eliminate double lits and tautologies
                ex = c[j] ^ x;
                if (ex == 0) { // double literals
                    delLitAt(c, j);
                    j--;
                    continue;
                } else if (ex == 1) { // tautology
                    setClSize(c, 0);
                    statsUpdate(p->stats, PP_EVNT_NUM_ELIM_CLS, 0);
                    break;
                }
            }
            x = c[j];
        }    
        assert(getClSize(c) == 0 || checkClSort(c));
    }

    // populate p->clst & p->cidx, p->clst contains clause addresses
    populate_aux(p, p->ocnf, p->clst);
    
    do {
        change = bcp(p);
        pure_lits(p);
        change |= bcp(p);
        if (p->sat_status != NONE) break;
        
        cmp_qsort(a, getFlexArraySize(p->ocnf), 16, clcmp);

        // do one-step resolution and count literal occurencies
        // the same as bcp, resolution can:
        // * detect conflicts
        // * remove lits from clauses and clause/eqvs aux_tables
        // * deletes double and one of resolved clauses
        // * sorts clauses and frees zero-sized clauses
        // * detects when all the clauses are satisfied
        // but can't satisfy the clause
        change |= onestep_resolve(p);
        if (p->sat_status != NONE) break;

        if (p->eqvPreprocessing) {
            (void)find_bieq(p);

            // find_equivalencies, removes clauses, but doesn't sort
            // or free them 
            change |= (change | first) ? eq_find(p) : FALSE;

            assert(bcpqEmpty(p->impq));

            // propagate_bieq, returns TRUE if there are any replacements 
            // and sorts cnf clauses if there's any change, but doesn't
            // free zero-sized clauses
            change |= propagate_bieq(p);
            if (p->sat_status != NONE) break;

            // reduces equivalencies, registers (and removes) tautologies
            // doesn't use cnf clauses at all
            change |= (change | first) ? eq_reason(p) : FALSE;
            if (p->sat_status != NONE) break;

            change |= (change) ? FALSE : eq_find_depvars(p);
            if (p->sat_status != NONE) break;
            
            change |= propagate_bieq(p);
            if (p->sat_status != NONE) break;
        }

        first = FALSE;
    } while (change);
    
    // update priorities for simplified CNF
    for (i = 1; i <= p->ovnum; i++) {
        uint l = i << 1;
        p->pri[l][PRI_SIMPLIFIED] = getFlexArraySize(p->clst[l]);
        l++;
        p->pri[l][PRI_SIMPLIFIED] = getFlexArraySize(p->clst[l]);
    }

    sz = getFlexArraySize(p->ocnf);
    for (i = 0; i < sz; i++) {
        cla_t *x = (cla_t *)a[i];
        if (isBinCl(x)) {
            p->stats->ppNumBinCls++;
            p->pri[x[1]][PRI_BIN_CLS] += 1;
            p->pri[x[2]][PRI_BIN_CLS] += 1;
        }
    }

    // add all dependant variables to depvars
    for (i = 0; i < p->eqs_ptr; i++) {
        cla_t *c = p->eqs[i];
		uint csz;
        assert(isEquiv(c));
        assert(checkClSort(c));
        csz = getClSize(c);
        if (csz == 0) continue;
        assert(csz > 3);
        assert(checkClSort(c));
        if (isTautology(c)) {
            appendToFlexArray(p->elimvars, c[csz]);
            statsUpdate(p->stats, PP_EVNT_TAUT, 0);
        } else {
            appendToFlexArray(p->depvars, c[csz]);
            appendToFlexArray(p->ecnf, (uint)c);
        }
    }
    
    if (p->clusterize) {
        cluster(p);
    }

    statsUpdate(p->stats, PP_EVNT_NUM_ELIM_VAR, 
            (long)getFlexArraySize(p->elimvars));
    statsUpdate(p->stats, PP_EVNT_NUMEQ, 
            (long)getFlexArraySize(p->depvars));

#ifdef DEBUG
    if (p->sat_status == NONE) {
        char *varr = xcalloc(1, (size_t)1 + p->ovnum);
        for (i = 0; i < p->eqs_ptr; i++) {
            cla_t *x = p->eqs[i];
            uint dep, sz = getClSize(x);
            if (sz == 0) continue;
            // dependant or tautology variable
            dep = x[sz];
            // must be a valid variable
            assert(dep > 0 && dep <= p->ovnum);
            // there can be only one taut/dep
            // clause with dep variable
            assert(varr[dep] == EQ_NONE);
            // dep/taut var must not be assigned
            // or replaced by some other variable
            assert(p->model[dep] == UINT_MAX);
            // must be present in the clause
            assert(findLit(x, dep) != -1);
            if (isTautology(x)) {
                assert(occurInCNF(p, dep) == 0);
                assert(occurInCOE(p, dep) == 0);
                varr[dep] = EQ_TAUTOLOGY;
            } else {
                assert(occurInCNF(p, dep) > 0);
                assert(occurInCOE(p, dep) == 1);
                varr[dep] = EQ_DEPENDANT;
            }
        }
        free(varr);
    }
#endif
}

void verify(preproc_t *p, char *instance) {
    uint dummy1, dummy2;
    flexarray_t *cls = parse(instance, &dummy1, &dummy2);
    uint *a, i, sz, *model = p->model;
    a = getFlexArray(cls);
    sz = getFlexArraySize(cls);
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        uint j, csz = getClSize(c);
        assert(csz > 1);
        for (j = 1; j < csz; j++) {
            uint lit = c[j];
            uint v = getVar(lit);
            assert(v > 0 && v <= p->ovnum);
            assert(model[v] < 2);
            // check if the literal is satisfied
            if (model[v] == LSB(lit)) {
                break;
            }
        }
        if (j == csz) {
            arrayClPrn(c);
            error("Solution incorrect - unsat clause found!");
        }
        free(c);
    }
    freeFlexArray(cls);
}


// if model is NULL, the problem is unsatisfiable
// if ofile is NULL, output will be sent to std out
void postprocess(preproc_t *p, flexarray_t *model, char format,
        char *ofile, char *instance, clock_t start) {
    double timer;
    FILE *fp = NULL;
    char *msg = xmalloc(256);
    uint i;
    
    if (p->sat_status == TRUE) {
        uint *a = getFlexArray(model);
        uint sz = getFlexArraySize(model);
        for (i = 0; i < sz; i++) {
            uint xl = a[i];
            uint xv = getVar(xl);
            assert(xv > 0 && xv <= p->ovnum);
            update_model(p, xv, getPhase(xl));
        }
        do {
            // solve tautologies
            while (!stackEmpty(p->taut_stack)) {
                cla_t *c = (cla_t *)stackPop(p->taut_stack);
                if (!c) break;
                resolveTautEqvCl(p, c);
            }

            // fill in equivalent variables (from binary equivalencies)
            for (i = 1; i <= p->ovnum; i++) { 
                uint lit = p->model[i];
                if (lit > 1 && lit < UINT_MAX) {
                    (void)resolveBinEq(p, i);
                }
            }
        } while (!stackEmpty(p->taut_stack));
    }

    // create output file if needed
    if (ofile) {
        fp = fopen(ofile, "a+");
        if (!fp) {
            error("Output file creation error!");
        }
    }

    sprintf(msg, "c Instance %s\n", instance);
    writeStream(msg, fp);
    timer = (double)(clock() - start) / CLOCKS_PER_SEC;
    sprintf(msg, "c TIME:Total run time: %f [s]\n", timer);
    writeStream(msg, fp);
    sprintf(msg, "c TIME:solver run time: %f [s]\n", timer - 
            p->stats->ppTimeCluster - p->stats->ppTimeGraphGen);
    writeStream(msg, fp);
    switch (p->sat_status) {
        case NONE:
            writeStream("s UNKNOWN\n", fp);
            writeStream("c Result not available (timeout?)...\n", fp);
            break;
        case TRUE:
            writeStream("s SATISFIABLE\n", fp);
            if (format == RAW) {
                writeStream("c Solution suppressed...\n", fp);
                goto after_dump;
            } else {
                writeStream("v ", fp);
                for (i = 1; i <= p->ovnum; i++) {
                    uint sgn = p->model[i];
                    assert(sgn < 2);
                    sprintf(msg, "%d ", (sgn) ? (int)i : -(int)i);
                    writeStream(msg, fp);
                }
                writeStream("0\n", fp);
            }
            break;
        case FALSE:
            writeStream("s UNSATISFIABLE\n", fp);
            break;
    }
after_dump:
    if (fp) fclose(fp);
    free(msg);
}
#endif

/*
 * Revision history:
 *
 * $Log: simplify.h,v $
 * Revision 1.9  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.8  2005/01/30 13:22:30  root
 * Fixed bug in SAT graph dump.
 *
 * Revision 1.7  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.5  2004/12/11 04:46:12  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.4  2004/12/09 18:54:47  root
 * Fixed a bug in onestep_resolve (skipping a clause 
 * with i++) and some other minor modifications.
 *
 * Revision 1.3  2004/10/31 19:13:29  root
 * Added literal check, fixed bugs in pure_lit and bcp 
 * (propagation of already replaced literals).
 *
 * Revision 1.2  2004/10/24 00:08:10  root
 * Fixed several bugs.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
