/* Author: Domagoj Babic 2004.
 *
 * $Id: sort.h,v 1.6 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  A part of this code is adapted from Non-Recursive ANSI quicksort
*  function, available at:
*  http://c.snippets.org/snip_lister.php?fname=rg_qsort.c
*  As this file could be considered as a derivative work, please see 
*  the webpage for licensing terms and copyright notice. 
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND Domagoj Babic MAKES NO 
*  REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF 
*  EXAMPLE, BUT NOT LIMITATION. Domagoj Babic MAKES NO 
*  REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS 
*  FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE 
*  WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS TRADEMARKS OR OTHER 
*  RIGHTS. Domagoj Babic SHALL NOT BE LIABLE FOR ANY LIABILITY OR 
*  DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY THIRD PARTY 
*  ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY SUBLICENSE OR 
*  USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Domagoj Babic shall not be liable under any circumstances for any 
*  direct, indirect, special, incidental, or consequential damages with 
*  respect to any claim by USER or any third party on account of or 
*  arising from the use, or inability to use, this software or its 
*  associated documentation, even if the Author has been advised of the 
*  possibility of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, I can't guarantee any support. The 
*  solver has been written exclusively for research and experimental 
*  purposes, so documentation and comments are scarce. Most likely, 
*  this will not change in the near future, but feel free to check my 
*  webpage occasionally, as new releases (if there will be any) will 
*  be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#ifndef H_SORT
#define H_SORT
// Quicksort will use insertion sort for arrays
// smaller than QSORT_LIMIT.
#define QSORT_LIMIT 8
#include <stddef.h>
#include "utils.h"

int binsearch(int tbl[], int length, int n);
inline void cmp_ssort(const void *basep, const uint len, 
        int (*cmp)(const void *, const void *));
//inline void cmp_isort(const void *basep, const uint len, 
//        int (*cmp)(const void *, const void *));
// compare functions
inline int ucmp(const void *a, const void *b);

inline void swp(uint *a, uint *b) {
    uint tmp = *a;
    *a = *b;
    *b = tmp;
}

inline int ucmp(const void *a, const void *b) {
    uint xa = *(uint *)a;
    uint xb = *(uint *)b;
    return (int)xa - xb;
}

#define SWAP(a, b) (swp(a, b))
#define COMP(a, b) ((*cmp)((void *)(a), (void *)(b)))

void cmp_qsort(const void *basep, const uint length, const int tresh,
        int (*cmp)(const void *, const void *)) {
    uint *stack[40], **sp;
    uint *i, *j, *right;
    ptrdiff_t treshold = (ptrdiff_t)((tresh > 2) ? tresh : 8);
    uint *left = (uint *)basep;
    sp = stack;
    right = left + length;
    for (;;) {
        ptrdiff_t tmp = (ptrdiff_t)(right - left);
        if (tmp > treshold) {
            SWAP((tmp >> 1) + left, left);
            i = left + 1;
            j = right - 1;
            if (COMP(i, j) > 0) SWAP(i, j);
            if (COMP(left, j) > 0) SWAP(left, j);
            if (COMP(i, left) > 0) SWAP(i, left);
            for (;;) {
                while(COMP(++i, left) < 0);
                while(COMP(--j, left) > 0);
                if (i > j) break;
                SWAP(i, j); 
            } 
            SWAP(left, j);
            if (j - left > right - i) {
                sp[0] = left;
                sp[1] = j;
                left = i;
            } else {
                sp[0] = i;
                sp[1] = right;
                right = j;
            }
            sp += 2;
        } else {
            cmp_ssort(left, (uint)tmp, cmp);
            if (sp != stack) {
                sp -= 2;
                left = sp[0];
                right = sp[1];
            } else break;
        }
    }
}

#define TCOMP(a, b) ((*cmp)(data, (void *)(a), (void *)(b)))

/* extended shell sort
void cmp_tssort(const void *basep, const uint len, const uint data,
        int (*cmp)(const uint, const void *, const void *)) {
    uint gap, i, j, *a, *b;
    // using Knuth's sequence 3*k + 1
    for (gap = 0; ++gap < len;) gap *= 3;
    while (gap /= 3) {
        for (i = gap; i < len; i++) {
            for (j = i - gap; ; j -= gap) {
                a = j + (uint *)basep;
                b = a + gap;
                if (TCOMP(a, b) <= 0) break;
                SWAP(a, b);
                if (j < gap) break;
            }
        }
    }
}// */

// insertion sort
inline void cmp_tisort(const void *basep, const uint len, const uint data,
        int (*cmp)(const uint, const void *, const void *)) {
    uint *i, *j, *left = (uint *)basep, *right = left + len;
    for (j = left, i = j + 1; i < right; j = i, i++) {
        for (; TCOMP(j, j + 1) > 0; j--) {
            SWAP(j, j + 1);
            if (j == left) break;
        }
    }
}

// extended quick sort
void cmp_tqsort(const void *basep, const uint length, const int tresh,
        const uint data,
        int (*cmp)(const uint, const void *, const void *)) {
    uint *stack[40], **sp;
    uint *i, *j, *right;
    ptrdiff_t treshold = (ptrdiff_t)((tresh > 2) ? tresh : 8);
    uint *left = (uint *)basep;
    sp = stack;
    right = left + length;
    for (;;) {
        ptrdiff_t tmp = (ptrdiff_t)(right - left);
        if (tmp > treshold) {
            SWAP((tmp >> 1) + left, left);
            i = left + 1;
            j = right - 1;
            if (TCOMP(i, j) > 0) SWAP(i, j);
            if (TCOMP(left, j) > 0) SWAP(left, j);
            if (TCOMP(i, left) > 0) SWAP(i, left);
            for (;;) {
                while(TCOMP(++i, left) < 0);
                while(TCOMP(--j, left) > 0);
                if (i > j) break;
                SWAP(i, j); 
            } 
            SWAP(left, j);
            if (j - left > right - i) {
                sp[0] = left;
                sp[1] = j;
                left = i;
            } else {
                sp[0] = i;
                sp[1] = right;
                right = j;
            }
            sp += 2;
        } else {
            cmp_tisort(left, (uint)tmp, data, cmp);
            if (sp != stack) {
                sp -= 2;
                left = sp[0];
                right = sp[1];
            } else break;
        }
    }
}

/* insertion sort
inline void cmp_isort(const void *basep, const uint len, 
        int (*cmp)(const void *, const void *)) {
    uint *i, *j, *left = (uint *)basep, *right = left + len;
    for (j = left, i = j + 1; i < right; j = i, i++) {
        for (; COMP(j, j + 1) > 0; j--) {
            SWAP(j, j + 1);
            if (j == left) break;
        }
    }
}// */

// shell sort
void cmp_ssort(const void *basep, const uint len,
        int (*cmp)(const void *, const void *)) {
    uint gap, i, j, *a, *b;
    // using Knuth's sequence 3*k + 1
    for (gap = 0; ++gap < len;) gap *= 3;
    while (gap /= 3) {
        for (i = gap; i < len; i++) {
            for (j = i - gap; ; j -= gap) {
                a = j + (uint *)basep;
                b = a + gap;
                if (COMP(a, b) <= 0) break;
                SWAP(a, b);
                if (j < gap) break;
            }
        }
    }
}

   
// for searching variables in clauses.
// preconditions:
// lit = (v << 1) + sign, sign = 1 | 0
// there's only one appearance of each var in clause
inline int exorsearch(int tbl[], const int length, const int n) {
    int low = 0;
    int high = length - 1;
    assert(low <= high);
    // WARRNING: this won't work if length is zero!!
    while (low <= high) { 
        int mid = (low + high) >> 1;
        if ((tbl[mid] ^ n) < 2) {
            return mid;
        } else if (n < tbl[mid]) {
            high = mid - 1;
        } else {
            low = mid + 1;
        }
    }
    return -1;
}

// returns the index of the element n, and -1
// it doesn't exist in the array
int binsearch(int tbl[], int length, int n) {
    int low=0;
    int high=length-1;
    int mid;

    while(low <= high){
        mid = (low + high) >> 1;
        if(n == tbl[mid]) return mid;
        else
            if(n < tbl[mid]) high=mid-1;
            else low=mid+1;
    }
    return -1;
}
#endif

/*
 * Revision history:
 *
 * $Log: sort.h,v $
 * Revision 1.6  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.5  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.3  2004/12/09 18:56:34  root
 * Added extended sort functions with additional 
 * parameter to be passed to CMP.
 *
 * Revision 1.2  2004/10/24 00:09:05  root
 * Modified exorsearch (to work faster with hypercubing).
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */

