/* Author : Domagoj Babic 2004.
 *
 * $Id: utils.h,v 1.7 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#ifndef H_UTILS
#define H_UTILS

#ifdef DEBUG
#define DBG(x) x
#else
#define NDEBUG
#define DBG(x)
#endif
#ifdef DUMP
#define DMP(x) x
#else
#define DMP(x)
#endif
#define DBGX(x) 
#define DMPX(x)

#ifdef __GNUC__
#define USE_MMAP
#endif

#ifdef DEBUG
#include <assert.h>
#else
#define assert(x)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef __GNUC__
#include <fcntl.h>
#include <unistd.h>
#endif

#ifdef USE_MMAP
#include <sys/mman.h>
#endif


#ifdef DMALLOC
#include "dmalloc.h"
#endif

#if !defined (__GNUC__) && !defined inline
#define inline __inline
#endif

#define error(str) report(__FILE__, __LINE__, str, ERROR)
#define warrn(str) report(__FILE__, __LINE__, str, WARRNING)
#define note(str) report(__FILE__, __LINE__, str, NOTE)
#define memOut() outOfMem(__FILE__, __LINE__)
#define uint unsigned int //uint_fast32_t
#define uintsh unsigned short int //uint_fast16_t
#define ulong unsigned long long int
#define uchar unsigned char
#define TRIM(x) (void *)((uint)x & (uint)(~3))
#define LSB(x) ((uint)x & (uint)1)
#define LSB2(x) ((uint)x & (uint)3)
#define TOBOOL(x) (x > 0)

#ifdef PAGELOCK
#define pagelock() mlockall(MCL_FUTURE|MCL_CURRENT)
#else
#define pagelock()
#endif

// boolean types extended with NONE (no specific value
// given). NONE is used for example in vartb->setDecision.
// If no value is specified (NONE), the function can select
// value on its own.
enum bool_e {FALSE = 0, TRUE, NONE};

// used primarly in debugging, for printing information about
// // failed pre/postconditions and obvious errors
enum severity_e {NOTE, WARRNING, ERROR};

// exit codes required by SAT competition
enum exit_codes_e {EXIT_SATISFIABLE = 10, EXIT_UNSATISFIABLE = 20,
    EXIT_UNKNOWN = 0};

void report(char *file, unsigned line, char *str, char sev) {
#ifdef DEBUG
    if (sev == ERROR) {
        printf("c %s:%u ERROR: %s \n", file, line, str);
        printf("s UNKNOWN\n");
        exit(EXIT_UNKNOWN);
    } else if (sev == WARRNING) {
        printf("c %s:%u WARRNING: %s \n", file, line, str);
    } else if (sev == NOTE) {
        printf("c %s:%u INFO: %s \n", file, line, str);
    }
#else
    if (sev == ERROR) {
        printf("c ERROR: %s \n", str);
        printf("s UNKNOWN\n");
        exit(EXIT_UNKNOWN);
    } else if (sev == WARRNING) {
        printf("c WARRNING: %s \n", str);
    } else if (sev == NOTE) {
        printf("c INFO: %s\n", str);
    }
#endif
}

void outOfMem(char *file, unsigned line) {
    report(file, line, "Out of memory error!", ERROR);
}

void writeStream(char *msg, FILE *f) {
    if (f) {
        fputs(msg, f);
    } else {
        printf("%s", msg);
    }
}

#ifdef USE_MMAP
void *xmmap(void *adr, size_t len, int prot, int flags, int fd,
        int offset) {
    char *p = mmap(adr, len, prot, flags, fd, offset);
    if (p == MAP_FAILED) {
        error("Unsuccessful MMAP call - possible causes:\n\
* your system doesn't support MAP_PRIVATE & MAP_ANON flags\n\
* system's out of memory\n");
    }
    return (void *)p;
}
#endif

void *xmalloc(size_t size) {
    void *ptr = malloc(size);
    if (!ptr) memOut();
    return ptr;
}

void *xcalloc(size_t nmemb, size_t size) {
    void *ptr = calloc(nmemb, size);
    if (!ptr) memOut();
    return ptr;
}

void *xrealloc(void *ptr, size_t size) {
    void *p = realloc(ptr, size);
    if (!p) memOut();
    return p;
}

#ifdef __GNUC__
// read an integer from random device
int devrand(void) {
    int rnd;
    int fd = open("/dev/urandom", O_RDONLY);
    if (fd == -1) error("/dev/urandom open error!");
    if (read(fd,&rnd,4) < 1)
        error("Could not read from /dev/urandom");
    close(fd);
    return rnd;
}
    
// returns pseudo random integer in range [0..limit-1]
int pseudorand(const int limit) {
    /*
    static char seeded = FALSE;
    if (!seeded) {
        // if not seeded, read a seed from random device
        srand((unsigned)devrand());
        seeded = TRUE;
    }
    // */
    return (int)(((float)limit)*rand()/RAND_MAX);
}
#endif

inline int _min(const int x, const int y) {
#ifdef UGLY_HACKS
    return (int)(y + ((x - y) & ((x - y) >> (sizeof(int) * 8 - 1))));
#else
    return (x < y) ? x : y;
#endif
}

inline uint _umin(const uint x, const uint y) {
    return (x < y) ? x : y;
}

inline int _max(const int x, const int y) {
#ifdef UGLY_HACKS
    return (int)(x - ((x - y) & ((x - y) >> (sizeof(int) * 8 - 1))));
#else
    return (x > y) ? x : y;
#endif
}

inline uint _umax(const uint x, const uint y) {
    return (x > y) ? x : y;
}

inline float maxf(const float x, const float y) {
    return (x > y) ? x : y;
}

#define LOG2_E 0.6931
#define LOG10_E 2.3026

inline uint xlog2(float x) {
    return (uint)((float)log(x) / LOG2_E);
}

inline uint xlog10(float x) {
    return (uint)((float)log(x) / LOG10_E);
}

uint approx_log2(uint v) {
    const unsigned int b[] = {(uint)0x2, (uint)0xC, 
        (uint)0xF0, (uint)0xFF00, (uint)0xFFFF0000};
    const unsigned int S[] = {1, 2, 4, 8, 16};
    int i;
    register unsigned int c = 0; // result of log2(v) will go here
    for (i = 4; i >= 0; i--) {
        if (v & b[i]) {
            v >>= S[i];
            c |= S[i];
        } 
    }
    return c;
}    

void swap(int s[], int i, int j){
    int tmp;
    tmp=s[i];
    s[i]=s[j];
    s[j]=tmp;
}

#endif // H_UTILS

/*
 * Revision history:
 *
 * $Log: utils.h,v $
 * Revision 1.7  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.6  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.4  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.3  2004/12/11 04:46:12  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.2  2004/12/09 18:55:23  root
 * Added a couple of new functions.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
